from datetime import datetime
from threading import Lock

from flask import Flask, render_template, request, jsonify
from flask_socketio import SocketIO, emit

app = Flask(__name__)
app.config['SECRET_KEY'] = 'your_secret_key_here'
socketio = SocketIO(app, cors_allowed_origins="*", async_mode='threading')

# 存储客户端信息
clients = {}  # 格式: {client_id: client_info}
socket_to_client = {}  # 映射: socket_sid -> client_id
client_lock = Lock()
terminal_sessions = {}
pending_connections = {}  # 等待终端连接的请求


@app.route('/')
def index():
    return render_template('index.html')


@app.route('/clients')
def get_clients():
    """获取客户端列表"""
    with client_lock:
        clients_list = [{
            'id': client_id,
            'name': info.get('name', 'Unknown'),
            'connected_at': info.get('connected_at'),
            'status': 'connected',
            'os': info.get('os')
        } for client_id, info in clients.items()]
    return jsonify(clients_list)


@socketio.on('client_register')
def handle_client_register(data):
    """客户端注册"""
    client_id = data.get('client_id')
    client_name = data.get('name', 'Unknown Client')
    client_os = data.get('os', 'Unknown Client')

    print(f"Client registering: {client_name} with ID: {client_id}")

    with client_lock:
        # 存储客户端信息
        clients[client_id] = {
            'id': client_id,
            'name': client_name,
            'connected_at': datetime.now().isoformat(),
            'socket_sid': request.sid,
            'type': 'client',
            'os': client_os
        }
        # 建立socket_sid到client_id的映射
        socket_to_client[request.sid] = client_id

    print(f"Client registered successfully: {client_name} (ID: {client_id})")
    print(f"Current clients: {list(clients.keys())}")

    emit('registration_success', {'status': 'registered'})

    # 检查是否有等待此客户端的终端连接请求
    if client_id in pending_connections:
        for frontend_sid in pending_connections[client_id]:
            start_terminal_session(client_id, frontend_sid)
        del pending_connections[client_id]

    # 广播客户端列表更新
    emit('clients_updated', broadcast=True)


def start_terminal_session(client_id, frontend_sid):
    """启动终端会话"""
    if client_id not in clients:
        return False

    # 建立终端会话映射
    terminal_sessions[client_id] = {
        'frontend_sid': frontend_sid,
        'client_sid': clients[client_id]['socket_sid']
    }

    # 通知前端终端已启动
    socketio.emit('terminal_started', {
        'client_id': client_id
    }, room=frontend_sid)

    print(f"Terminal session started for client: {client_id}")
    return True


@socketio.on('terminal_data')
def handle_terminal_data(data):
    """处理终端数据转发"""
    client_id = data.get('client_id')
    output = data.get('data')

    # 将客户端终端的输出转发给前端
    if client_id in terminal_sessions:
        emit('terminal_output', {
            'data': output,
            'client_id': client_id
        }, room=terminal_sessions[client_id]['frontend_sid'])


@socketio.on('frontend_connect')
def handle_frontend_connect():
    """前端连接"""
    print("Frontend connected")


@socketio.on('start_terminal')
def handle_start_terminal(data):
    """启动终端会话"""
    client_id = data.get('client_id')
    frontend_sid = request.sid

    print(f"Starting terminal for client: {client_id}")
    print(f"Available clients: {list(clients.keys())}")

    # 检查客户端是否存在
    if client_id in clients:
        # 客户端已注册，直接启动终端会话
        if start_terminal_session(client_id, frontend_sid):
            emit('terminal_status', {
                'status': 'connected',
                'client_id': client_id,
                'message': '终端连接已建立'
            })
        else:
            emit('terminal_error', {'error': '无法启动终端会话'})
    else:
        # 客户端尚未注册，将请求加入等待队列
        with client_lock:
            if client_id not in pending_connections:
                pending_connections[client_id] = []
            pending_connections[client_id].append(frontend_sid)

        emit('terminal_status', {
            'status': 'waiting',
            'client_id': client_id,
            'message': '等待客户端连接...'
        })
        print(f"Waiting for client {client_id} to register")


@socketio.on('terminal_input')
def handle_terminal_input(data):
    """处理前端终端输入"""
    client_id = data.get('client_id')
    input_data = data.get('data')

    # 将前端的输入转发给对应的客户端
    if client_id in terminal_sessions:
        client_sid = terminal_sessions[client_id]['client_sid']
        emit('terminal_command', {
            'command': input_data,
            'client_id': client_id
        }, room=client_sid)


@socketio.on('disconnect')
def handle_disconnect():
    """处理连接断开"""
    print(f"Client disconnected: {request.sid}")

    with client_lock:
        # 如果是客户端断开连接
        if request.sid in socket_to_client:
            client_id = socket_to_client[request.sid]

            # 清理客户端信息
            if client_id in clients:
                clients.pop(client_id)

            socket_to_client.pop(request.sid)

            # 清理终端会话
            if client_id in terminal_sessions:
                terminal_sessions.pop(client_id)

            print(f"Removed client: {client_id}")

        # 如果是前端断开连接，清理对应的终端会话
        for client_id in list(terminal_sessions.keys()):
            if terminal_sessions[client_id]['frontend_sid'] == request.sid:
                terminal_sessions.pop(client_id)
                print(f"Removed terminal session for client: {client_id}")

        # 清理等待队列中的前端请求
        for client_id in list(pending_connections.keys()):
            if request.sid in pending_connections[client_id]:
                pending_connections[client_id].remove(request.sid)
                if not pending_connections[client_id]:
                    del pending_connections[client_id]

    emit('clients_updated', broadcast=True)


def start():
    socketio.run(app, host='0.0.0.0', port=5000, debug=True, allow_unsafe_werkzeug=True)


if __name__ == '__main__':
    start()



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: qicongsheng
from setuptools import setup, find_packages

from mecho import help

setup(
    name="mecho",
    version=help.get_version(),
    author="qicongsheng",
    author_email="qicongsheng@outlook.com",
    description="mecho",
    url="https://github.com/qicongsheng/mecho",
    packages=find_packages(),
    package_data={
        "mecho": ["templates/*", "static/*", "static/*/*"]
    },
    install_requires=[
        "flask==2.3.3",
        "flask-socketio==5.3.6",
        "python-socketio[client]==5.10.0",
        "requests==2.31.0"
    ],
    entry_points={
        "console_scripts": [
            "mecho-server=mecho.server:start",
            "mecho-client=mecho.client:start",
        ]
    }
)



import os
import threading
import time
import uuid

import select
import socketio

# 检查平台并导入相应的模块
if os.name == 'nt':  # Windows
    import subprocess

    PTY_AVAILABLE = False
else:  # Unix/Linux/MacOS
    try:
        import pty
        import termios
        import tty

        PTY_AVAILABLE = True
    except ImportError:
        PTY_AVAILABLE = False


class TerminalClient:
    def __init__(self, server_url, client_name=None):
        self.sio = socketio.Client()
        self.server_url = server_url
        # 在初始化时就生成client_id，确保一致性
        self.client_id = str(uuid.uuid4())
        self.client_name = client_name or f"{self.get_hostname()}"
        self.master_fd = None
        self.process = None
        self.running = False
        self.connected = False

        self.sessions = {}
        self.os = os.name

        print(f"初始化客户端: {self.client_name}")
        print(f"系统平台: {os.name}")
        print(f"PTY 可用: {PTY_AVAILABLE}")

        self.setup_handlers()

    def get_hostname(self):
        """获取主机名 - 跨平台兼容"""
        try:
            import socket
            return socket.gethostname()
        except:
            return "unknown-host"

    def setup_handlers(self):
        @self.sio.event
        def connect():
            print(f"✓ 连接到服务器: {self.server_url}")
            self.connected = True
            # 立即注册客户端
            self.sio.emit('client_register', {
                'client_id': self.client_id,
                'name': self.client_name,
                'os': self.os
            })
            self.sessions[self.client_id] = {'input_buffer': ''}
            print(f"✓ 客户端注册信息已发送: {self.client_name} (ID: {self.client_id})")

        @self.sio.event
        def disconnect():
            print("✗ 与服务器断开连接")
            self.connected = False
            self.stop_terminal()

        @self.sio.on('registration_success')
        def handle_registration_success(data):
            print("✓ 客户端注册成功")

        @self.sio.on('terminal_command')
        def handle_terminal_command(data):
            """处理来自服务器的终端命令"""
            if data.get('client_id') != self.client_id:
                return

            command = data.get('command', '')
            print(f"收到终端输入: {repr(command)}")

            # Windows 和 Unix 不同的处理方式
            if os.name == 'nt':  # Windows
                session = self.sessions[self.client_id]
                if self.process and self.process.poll() is None:
                    try:
                        # 如果是回车键，发送整个命令并清空缓冲区
                        if command == '\r' or command == '\n':
                            command = session['input_buffer'] + '\r\n'
                            self.process.stdin.write(command)
                            self.process.stdin.flush()
                            session['input_buffer'] = ''  # 清空缓冲区
                        else:
                            # 将字符添加到缓冲区
                            session['input_buffer'] += command
                        print(f"✓ 成功写入命令到 Windows 进程")
                    except Exception as e:
                        print(f"✗ 写入 Windows 进程失败: {e}")
            else:  # Unix/Linux/MacOS
                if self.master_fd and self.running:
                    try:
                        os.write(self.master_fd, command.encode('utf-8'))
                        print(f"✓ 成功写入命令到伪终端")
                    except Exception as e:
                        print(f"✗ 写入伪终端失败: {e}")

    def find_available_shell(self):
        """查找可用的 shell - 跨平台兼容"""
        if os.name == 'nt':  # Windows
            # Windows 使用 cmd.exe 或 PowerShell
            shells = ['cmd.exe', 'powershell.exe']
            for shell in shells:
                # 检查是否在 PATH 中
                if self.which(shell):
                    print(f"找到可用 shell: {shell}")
                    return shell
            return 'cmd.exe'  # 默认回退
        else:  # Unix/Linux/MacOS
            shells = [
                '/bin/bash', '/usr/bin/bash',
                '/bin/zsh', '/usr/bin/zsh',
                '/bin/sh', '/usr/bin/sh',
                '/bin/ksh', '/usr/bin/ksh',
                '/bin/tcsh', '/usr/bin/tcsh'
            ]

            for shell in shells:
                if os.path.exists(shell):
                    print(f"找到可用 shell: {shell}")
                    return shell

            # 如果找不到任何标准 shell，尝试使用环境变量中的 SHELL
            fallback_shell = os.environ.get('SHELL')
            if fallback_shell and os.path.exists(fallback_shell):
                print(f"使用环境变量中的 shell: {fallback_shell}")
                return fallback_shell

            print("警告: 未找到可用的 shell，将尝试使用 /bin/sh")
            return '/bin/sh'  # 最后尝试

    def which(self, program):
        """查找可执行程序 - 类似 Unix which 命令的 Windows 实现"""

        def is_exe(fpath):
            return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

        fpath, fname = os.path.split(program)
        if fpath:
            if is_exe(program):
                return program
        else:
            for path in os.environ["PATH"].split(os.pathsep):
                exe_file = os.path.join(path, program)
                if is_exe(exe_file):
                    return exe_file
        return None

    def start_terminal(self):
        """启动终端 - 跨平台兼容版本"""
        try:
            print("正在启动终端...")

            if os.name == 'nt':  # Windows
                return self.start_windows_terminal()
            else:  # Unix/Linux/MacOS
                return self.start_unix_terminal()

        except Exception as e:
            print(f"✗ 启动终端失败: {e}")
            import traceback
            traceback.print_exc()
            return False

    def start_windows_terminal(self):
        """Windows 终端实现"""
        try:
            shell_path = self.find_available_shell()
            print(f"使用 Windows shell: {shell_path}")

            # 启动进程
            self.process = subprocess.Popen(
                [shell_path],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                text=True,
                bufsize=1,
                universal_newlines=True,
                cwd=os.getcwd()
            )

            self.running = True

            # 启动线程读取输出
            output_thread = threading.Thread(target=self.read_windows_output)
            output_thread.daemon = True
            output_thread.start()

            print("✓ Windows 终端启动成功")
            return True

        except Exception as e:
            print(f"✗ 启动 Windows 终端失败: {e}")
            return False

    def start_unix_terminal(self):
        """Unix/Linux/MacOS 终端实现"""
        if not PTY_AVAILABLE:
            print("✗ pty 模块不可用，无法创建伪终端")
            return False

        try:
            # 查找可用的 shell
            shell_path = self.find_available_shell()
            shell_name = os.path.basename(shell_path)

            print(f"使用 Unix shell: {shell_path}")

            # 创建伪终端
            self.master_fd, slave_fd = pty.openpty()

            # 启动shell进程
            pid = os.fork()

            if pid == 0:  # 子进程
                os.close(self.master_fd)
                os.setsid()
                os.dup2(slave_fd, 0)  # stdin
                os.dup2(slave_fd, 1)  # stdout
                os.dup2(slave_fd, 2)  # stderr
                os.close(slave_fd)

                # 设置环境变量
                env = os.environ.copy()
                env['TERM'] = 'xterm-256color'
                env['COLORTERM'] = 'truecolor'

                # 启动shell
                if shell_name in ['bash', 'zsh', 'ksh']:
                    os.execve(shell_path, [shell_name, '-i'], env)
                else:
                    os.execve(shell_path, [shell_name], env)
            else:
                os.close(slave_fd)
                self.running = True

                # 启动线程读取终端输出
                output_thread = threading.Thread(target=self.read_unix_output)
                output_thread.daemon = True
                output_thread.start()

                print("✓ Unix 终端启动成功")
                return True

        except Exception as e:
            print(f"✗ 启动 Unix 终端失败: {e}")
            return False

    def read_windows_output(self):
        """读取 Windows 进程输出"""
        print("开始读取 Windows 进程输出...")
        while self.running and self.process and self.process.poll() is None:
            try:
                output = self.process.stdout.readline()
                if output and self.connected and output != '\n':
                    self.sio.emit('terminal_data', {
                        'client_id': self.client_id,
                        'data': '\n' + output[:-1]
                    })
                    print(f"发送 Windows 输出: {repr(output[:50])}")
            except Exception as e:
                if self.running:
                    print(f"读取 Windows 输出时出错: {e}")
                break
        print("Windows 输出读取线程结束")

    def read_unix_output(self):
        """读取 Unix 伪终端输出"""
        print("开始读取 Unix 终端输出...")

        while self.running and self.master_fd:
            try:
                # 使用select检查是否有数据可读
                r, w, e = select.select([self.master_fd], [], [], 0.1)
                if self.master_fd in r:
                    output = os.read(self.master_fd, 1024).decode('utf-8', errors='ignore')
                    if output and self.connected:
                        # 发送终端输出到服务器
                        self.sio.emit('terminal_data', {
                            'client_id': self.client_id,
                            'data': output
                        })
                        print(f"发送 Unix 输出: {repr(output[:50])}")
            except (OSError, ValueError) as e:
                if self.running:  # 只在运行状态下打印错误
                    print(f"读取 Unix 终端输出错误: {e}")
                break

        print("Unix 输出读取线程结束")

    def stop_terminal(self):
        """停止终端 - 跨平台兼容"""
        print("正在停止终端...")
        self.running = False

        # Windows 进程处理
        if self.process:
            try:
                if os.name == 'nt':
                    # Windows 发送 Ctrl+C 信号
                    self.process.send_signal(subprocess.signal.CTRL_C_EVENT)
                else:
                    self.process.terminate()

                self.process.wait(timeout=3)
            except:
                try:
                    self.process.kill()
                except:
                    pass
            self.process = None

        # Unix 伪终端处理
        if self.master_fd:
            try:
                os.close(self.master_fd)
            except:
                pass
            self.master_fd = None

        print("终端已停止")

    def connect_with_retry(self, max_retries=5, retry_interval=3):
        """带重试的连接"""
        for attempt in range(max_retries):
            try:
                print(f"尝试连接到服务器 ({attempt + 1}/{max_retries})...")
                self.sio.connect(self.server_url)
                self.start_terminal()
                return True
            except Exception as e:
                print(f"连接失败: {e}")
                if attempt < max_retries - 1:
                    print(f"{retry_interval}秒后重试...")
                    time.sleep(retry_interval)

        return False

    def run(self):
        """运行客户端"""
        if self.connect_with_retry():
            try:
                print("客户端运行中，按 Ctrl+C 退出...")
                self.sio.wait()
            except KeyboardInterrupt:
                print("\n客户端正在关闭...")
            except Exception as e:
                print(f"客户端运行错误: {e}")
            finally:
                self.stop_terminal()
                if self.sio.connected:
                    self.sio.disconnect()
        else:
            print("无法连接到服务器，客户端退出")


def start():
    import argparse

    parser = argparse.ArgumentParser(description='终端客户端')
    parser.add_argument('--name', type=str, help='客户端名称', default=os.getenv("NAME"))
    parser.add_argument('--server', type=str, help='服务器地址',
                        default=os.getenv("SERVER_URL", 'http://localhost:5000'))

    args = parser.parse_args()

    client = TerminalClient(
        server_url=args.server,
        client_name=args.name
    )

    client.run()


if __name__ == '__main__':
    start()
